from datetime import datetime
from cryptomodel.coinmarket import prices
from cryptomodel.fixer import exchange_rates
from cryptomodel.readonly import SymbolRates
from mongoengine import Q
from cryptodataaccess.helpers import server_time_out_wrapper, do_connect

DATE_FORMAT = "%Y-%m-%d"


class RatesRepository:

    def __init__(self, config, log_error):
        self.configuration = config
        self.log_error = log_error

    def fetch_symbols(self):
        symbols = {}
        latest_prices = self.fetch_latest_prices_to_date(datetime.today().strftime(DATE_FORMAT))
        for coin in latest_prices[0].coins:
            symbols.update({coin.symbol: coin.name})
        return symbols

    def fetch_symbol_rates(self):
        dt_now = datetime.today().strftime(DATE_FORMAT)
        srs = SymbolRates(dt_now)
        latest_prices = self.fetch_latest_prices_to_date(dt_now)
        for coin in latest_prices[0].coins:
            srs.add_rate(coin.symbol, coin.quote.eur)
        return srs

    def fetch_latest_prices_to_date(self, before_date):
        return server_time_out_wrapper(self, self.do_fetch_latest_prices_to_date, before_date)

    def fetch_latest_exchange_rates_to_date(self, before_date):
        return server_time_out_wrapper(self, self.do_fetch_latest_exchange_rates_to_date, before_date)

    def do_fetch_latest_prices_to_date(self, before_date):
        do_connect(self.configuration)
        return prices.objects(Q(status__timestamp__lte=before_date)).order_by(
            '-status__timestamp')[:10]

    def do_fetch_latest_exchange_rates_to_date(self, before_date):
        do_connect(self.configuration)
        return exchange_rates.objects(Q(date__lte=before_date)).order_by(
            'date-')[:1]



from cryptomodel.cryptostore import user_notification, user_channel, user_transaction, operation_type
from mongoengine import Q
from cryptodataaccess import helpers
from cryptodataaccess.helpers import if_none_raise, if_none_raise_with_id


class TransactionRepository:

    def __init__(self, config, log_error):
        self.configuration = config
        self.log_error = log_error

    def fetch_distinct_user_ids(self):
        return helpers.server_time_out_wrapper(self, self.do_fetch_transaction, id)

    def fetch_transaction(self, id):
        return helpers.server_time_out_wrapper(self, self.do_fetch_transaction, id)

    def fetch_transactions(self, user_id):
        return helpers.server_time_out_wrapper(self, self.do_fetch_transactions, user_id)

    def fetch_transactions(self, user_id):
        return helpers.server_time_out_wrapper(self, self.do_fetch_transactions, user_id)

    def fetch_distinct_user_ids(self):
        return helpers.server_time_out_wrapper(self, self.do_fetch_distinct_user_ids())

    def insert_transaction(self, user_id, volume, symbol, value, price, currency, date, source, source_id, operation):
        return helpers.server_time_out_wrapper(self, self.do_insert_transaction, user_id, volume, symbol,
                                               value, price, currency, date, source, source_id, operation)

    def update_transaction(self, id, user_id, volume, symbol, value, price, currency, date, source, source_id,
                           operation):
        return helpers.server_time_out_wrapper(self, self.do_update_transaction, id,
                                               user_id, volume, symbol, value, price, currency, date, source, source_id,
                                               operation)

    def delete_transaction(self, id, throw_if_does_not_exist=True):
        helpers.server_time_out_wrapper(self, self.do_delete_transaction, id, throw_if_does_not_exist)

    def do_delete_transaction(self, id, throw_if_does_not_exist=True):
        helpers.do_connect(self.configuration)
        trans = user_transaction.objects(id=id).first()
        if throw_if_does_not_exist:
            if_none_raise_with_id(id, trans)
        if trans is not None:
            trans.delete()

    def delete_transaction_by_source_id(self, source_id, throw_if_does_not_exist=True):
        helpers.server_time_out_wrapper(self, self.do_delete_transaction, source_id, throw_if_does_not_exist)

    def do_delete_transaction_by_source_id(self, source_id, throw_if_does_not_exist=True):
        helpers.do_connect(self.configuration)
        trans = user_transaction.objects(source_id=source_id).first()
        if throw_if_does_not_exist:
            if_none_raise_with_id(id, trans)
        if trans is not None:
            trans.delete()

    def do_update_transaction(self, id, user_id, volume, symbol, value, price, currency, date, source, source_id,
                              operation):
        helpers.do_connect(self.configuration)
        trans = user_transaction.objects(id=id).first()
        if_none_raise_with_id(id, trans)
        trans.user_id = user_id
        trans.volume = volume
        trans.symbol = symbol
        trans.value = value
        trans.price = price
        trans.date = date
        trans.source = source
        trans.currency = currency
        trans.source_id = source_id
        trans.operation = operation
        trans.save()
        return user_transaction.objects(id=id).first()

    def do_insert_transaction(self, user_id, volume, symbol, value, price, currency, date, source, source_id,
                              operation):
        helpers.do_connect(self.configuration)
        trans = user_transaction()
        trans.user_id = user_id
        trans.volume = volume
        trans.symbol = symbol
        trans.value = value
        trans.price = price
        trans.date = date
        trans.currency = currency
        trans.source = source
        trans.source_id = source_id
        trans.operation = operation
        trans.save()
        return user_transaction.objects(id=trans.id).first()

    def do_fetch_transactions(self, user_id ):
        helpers.do_connect(self.configuration)
        return user_transaction.objects(Q(user_id=user_id))

    def do_fetch_transaction(self,id):
        helpers.do_connect(self.configuration)
        return user_transaction.objects(Q(id=id))[0]

    def do_fetch_distinct_user_ids(self):
        helpers.do_connect(self.configuration)
        return user_transaction.objects().only('user_id').distinct('user_id')



import mock
from bson import ObjectId
from pymongo.errors import ServerSelectionTimeoutError
from cryptomodel.cryptostore import user_channel, user_transaction, user_notification
from cryptomodel.coinmarket import prices
from cryptomodel.fixer import exchange_rates
from cryptodataaccess.config import configure_app
from cryptodataaccess.UsersRepository import UsersRepository
from cryptodataaccess.RatesRepository import RatesRepository
import pytest
from cryptodataaccess.helpers import do_connect
from tests.helpers import insert_prices_record, insert_exchange_record


@pytest.fixture(scope='module')
def mock_log():
    with mock.patch("cryptodataaccess.helpers.log_error"
                    ) as _mock:
        _mock.return_value = True
        yield _mock


def test_fetch_symbol_rates():
    config = configure_app()
    repo = RatesRepository(config, mock_log)
    do_connect(config)
    prices.objects.all().delete()
    insert_prices_record()
    objs = repo.fetch_symbol_rates()
    assert (len(objs.rates) == 100)
    assert (objs.rates['BTC'].price == 8101.799293468747)


def test_fetch_exchange_rates():
    config = configure_app()
    repo = RatesRepository(config, mock_log)
    do_connect(config)
    exchange_rates.objects.all().delete()
    insert_exchange_record()
    objs = repo.fetch_latest_exchange_rates_to_date('1900-01-01')
    assert (len(objs) == 0)
    objs = repo.fetch_latest_exchange_rates_to_date('2020-07-04')
    assert (len(objs) == 1)
    objs = repo.fetch_latest_exchange_rates_to_date('2020-07-03')
    assert (len(objs) == 1)
    assert (objs[0].rates.AED == 4.127332)
    objs = repo.fetch_latest_exchange_rates_to_date('2020-07-02')
    assert (len(objs) == 0)


def test_fetch_prices_and_symbols():
    config = configure_app()
    repo = RatesRepository(config, mock_log)
    do_connect(config)
    prices.objects.all().delete()
    insert_prices_record()
    objs = repo.fetch_latest_prices_to_date('2020-07-03')  # bound case : timestamp is saved as string so it cant find
    # it because  of the time stuff (either+1 day?)
    assert (len(objs) == 0)
    objs = repo.fetch_latest_prices_to_date('2020-07-04')
    assert (len(objs) == 1)
    symbols = repo.fetch_symbols()
    assert (len(symbols) == 100)


def test_insert_user_channel():
    config = configure_app()
    repo = UsersRepository(config, mock_log)
    do_connect(config)
    user_channel.objects.all().delete()
    uc = repo.insert_user_channel(1, 'da', '1')
    assert (uc.channel_type == 'da')


def test_log_when_do_connect_raises_exception(mock_log):
    with mock.patch("cryptodataaccess.helpers.do_connect"
                    ) as _mock:
        _mock.side_effect = ServerSelectionTimeoutError("hi")
        with mock.patch("cryptodataaccess.helpers.log_error") as log:
            with pytest.raises(ServerSelectionTimeoutError):
                repo = UsersRepository(configure_app(), mock_log)
                repo.insert_user_channel(1, "telegram", chat_id="1")
            mock_log.assert_called()


def test_update_notification_when_does_not_exist_throws_ValueError():
    config = configure_app()
    repo = UsersRepository(config, mock_log)
    do_connect(config)
    user_notification.objects.all().delete()
    with pytest.raises(ValueError):
        repo.update_notification(ObjectId('666f6f2d6261722d71757578'), 1, 'nik2', 'email', 'some expr', 1, 1, True,
                                 'telegram', 'expr to send', ObjectId('666f6f2d6261722d71757578'), 'Modified')


def test_update_notification():
    config = configure_app()
    repo = UsersRepository(config, mock_log)
    do_connect(config)
    user_notification.objects.all().delete()
    un = repo.insert_notification(1, 'username', 'email', 'some expr', 1, 1, True, 'telegram', 'expr to send',
                                  ObjectId('666f6f2d6261722d71757578'), 'Added')
    un = repo.update_notification(un.id, 1, 'nik2', 'email', 'some expr', 1, 1, True, 'telegram', 'expr to send',
                                  ObjectId('666f6f2d6261722d71757578'), 'Modified')
    assert (un.user_name == "nik2")


def test_delete_notification_when_exists():
    config = configure_app()
    repo = UsersRepository(config, mock_log)
    do_connect(config)
    user_notification.objects.all().delete()
    ut = repo.insert_notification(1, 'username', 'email', 'some expr', 1, 1, True, 'telegram', 'expr to send',
                                  ObjectId('666f6f2d6261722d71757578'), 'Added')
    assert (len(user_notification.objects) == 1)
    ut = repo.delete_notification(ut.id)
    assert (len(user_notification.objects) == 0)


def test_delete_user_notification_when_exists_by_source_id():
    config = configure_app()
    repo = UsersRepository(config, mock_log)
    do_connect(config)
    un = repo.insert_notification(1, 'username', 'email', 'some expr', 1, 1, True, 'telegram', 'expr to send',
                                  ObjectId('666f6f2d6261722d71757578'), 'Added')
    assert (len(user_notification.objects) == 1)
    ut = repo.do_delete_user_notification_by_source_id(source_id=ObjectId('666f6f2d6261722d71757578'))
    assert (len(user_notification.objects) == 0)



from mongoengine import Q
from cryptomodel.cryptostore import user_notification, user_channel, user_transaction, operation_type
from cryptomodel.cryptostore import user_settings
from cryptodataaccess import helpers
from cryptodataaccess.helpers import if_none_raise, if_none_raise_with_id

DATE_FORMAT = "%Y-%m-%d"


class UsersRepository:

    def __init__(self, config, log_error):
        self.configuration = config
        self.log_error = log_error

    def fetch_user_channels(self, user_id):
        return helpers.server_time_out_wrapper(self, self.do_fetch_user_channels, user_id)

    def fetch_notifications(self, items_count):
        return helpers.server_time_out_wrapper(self, self.do_fetch_notifications, items_count)

    def insert_notification(self, user_id, user_name, user_email, expression_to_evaluate, check_every_seconds,
                            check_times, is_active, channel_type, fields_to_send,source_id, operation):
        return helpers.server_time_out_wrapper(self, self.do_insert_notification, user_id, user_name, user_email,
                                               expression_to_evaluate, check_every_seconds,
                                               check_times, is_active, channel_type, fields_to_send,source_id, operation)

    def update_notification(self,id, user_id, user_name, user_email, expression_to_evaluate, check_every_seconds,
                            check_times, is_active, channel_type, fields_to_send,source_id, operation):
        return helpers.server_time_out_wrapper(self, self.do_update_notification, id, user_id, user_name, user_email,
                                               expression_to_evaluate, check_every_seconds,
                                               check_times, is_active, channel_type, fields_to_send,source_id, operation)

    def update_user_settings(self, user_id, preferred_currency):
        return helpers.server_time_out_wrapper(self, self.do_update_user_settings, user_id, preferred_currency)

    def insert_user_settings(self, user_id, preferred_currency):
        return helpers.server_time_out_wrapper(self, self.do_insert_user_channel, user_id, preferred_currency)

    def insert_user_channel(self, user_id, channel_type, chat_id):
        return helpers.server_time_out_wrapper(self, self.do_insert_user_channel, user_id, channel_type, chat_id)

    def delete_notification(self, id):
        helpers.server_time_out_wrapper(self, self.do_delete_notification, id)

    def delete_user_settings(self, id):
        helpers.server_time_out_wrapper(self, self.do_delete_user_settings, id)

    def do_fetch_user_channels(self, user_id):
        helpers.do_connect(self.configuration)
        return user_notification.objects(Q(user_id=user_id))

    def do_fetch_notifications(self, items_count):
        helpers.do_connect(self.configuration)
        return user_notification.objects()[:items_count]

    def do_fetch_transactions(self, user_id):
        helpers.do_connect(self.configuration)
        return user_transaction.objects(Q(user_id=user_id))

    def do_insert_notification(self, user_id, user_name, user_email, expression_to_evaluate, check_every_seconds,
                               check_times, is_active, channel_type, fields_to_send,source_id, operation):
        helpers.do_connect(self.configuration)
        un = user_notification()
        un.user_id = user_id
        un.user_name = user_name
        un.user_email = user_email
        un.is_active = True
        un.expression_to_evaluate = expression_to_evaluate
        un.check_every_seconds = check_every_seconds
        un.check_times = check_times
        un.is_active = is_active
        un.channel_type = channel_type
        un.fields_to_send = fields_to_send
        un.source_id = source_id
        un.operation = operation
        un.save()
        return user_notification.objects(id=un.id).first()

    def do_insert_user_channel(self, user_id, channel_type, chat_id):
        helpers.do_connect(self.configuration)
        uc = user_channel()
        uc.userId = user_id
        uc.channel_type = channel_type
        uc.chat_id = chat_id
        uc.save()
        return user_channel.objects(id=uc.id).first()

    def do_insert_user_settings(self, user_id, preferred_currency):
        helpers.do_connect(self.configuration)
        us = user_settings()
        us.userId = user_id
        us.preferred_currency = preferred_currency
        us.save()
        return user_settings.objects(id=us.id).first()

    def do_update_user_settings(self, id, user_id, preferred_currency):
        helpers.do_connect(self.configuration)
        us = user_settings.objects(id=id).first()
        if_none_raise_with_id(id, us)
        us.user_id = user_id
        us.preferred_currency = preferred_currency
        us.save()
        return user_settings.objects(id=id).first()

    def do_update_notification(self, id, user_id, user_name, user_email, expression_to_evaluate, check_every_seconds,
                               check_times, is_active, channel_type, fields_to_send, source_id, operation):
        helpers.do_connect(self.configuration)
        un = user_notification.objects(id=id).first()
        if_none_raise_with_id(id, un)
        un.id = id
        un.userId = user_id
        un.user_name = user_name
        un.user_email = user_email
        un.is_active = True
        un.expression_to_evaluate = expression_to_evaluate
        un.check_every_seconds = check_every_seconds
        un.check_times = check_times
        un.is_active = is_active
        un.channel_type = channel_type
        un.fields_to_send = fields_to_send
        un.source_id = source_id
        un.operation = operation
        un.save()
        return user_notification.objects(id=un.id).first()

    def do_delete_notification(self, id):
        helpers.do_connect(self.configuration)
        un = user_notification.objects(id=id).first()
        if_none_raise_with_id(id, un)
        un.delete()

    def do_delete_user_settings(self, id):
        helpers.do_connect(self.configuration)
        us = user_settings.objects(id=id).first()
        if_none_raise_with_id(id, us)
        us.delete()

    def delete_user_notification_by_source_id(self, source_id, throw_if_does_not_exist=True):
        helpers.server_time_out_wrapper(self, self.do_delete_user_notification_by_source_id, source_id, throw_if_does_not_exist)

    def do_delete_user_notification_by_source_id(self, source_id, throw_if_does_not_exist=True):
        helpers.do_connect(self.configuration)
        un = user_notification.objects(source_id=source_id).first()
        if throw_if_does_not_exist:
            if_none_raise_with_id(id, un)
        if un is not None:
            un.delete()


import mock
from bson import ObjectId
from cryptomodel.cryptostore import user_channel, user_transaction, user_notification
from cryptodataaccess.config import configure_app
from cryptodataaccess.TransactionRepository import TransactionRepository
import pytest
from cryptodataaccess.helpers import do_connect


@pytest.fixture(scope='module')
def mock_log():
    with mock.patch("cryptodataaccess.helpers.log_error"
                    ) as _mock:
        _mock.return_value = True
        yield _mock


def test_insert_transaction():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    user_transaction.objects.all().delete()
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    assert (ut.user_id == 1)
    assert (ut.symbol == "OXT")
    assert (len(user_transaction.objects) == 1)
    assert (ut.operation == 'Added')


def test_update_transaction():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    user_transaction.objects.all().delete()
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    ut = repo.update_transaction(ut.id, 1, 1, 'OXT2', 1, 1, "EUR", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    assert (ut.user_id == 1)
    assert (ut.symbol == "OXT2")
    assert (ut.currency == "EUR")
    assert (ut.operation == 'Added')


def test_update_transaction_when_does_not_exist_throws_ValueError():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    user_transaction.objects.all().delete()
    with pytest.raises(ValueError):
        repo.update_transaction(ObjectId('666f6f2d6261722d71757578'), 1, 1, 'OXT', "EUR", 1, 1, "2020-01-01",
                                "kraken", source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')


def test_delete_transaction_when_does_not_exist_throws_ValueError():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    user_transaction.objects.all().delete()
    with pytest.raises(ValueError):
        repo.delete_transaction(ObjectId('666f6f2d6261722d71757578'))


def test_delete_transaction_when_does_not_exist_and_throw_is_false_does_not_throw():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    user_transaction.objects.all().delete()
    repo.delete_transaction(ObjectId('666f6f2d6261722d71757578'), throw_if_does_not_exist=False)


def test_delete_transaction_when_exists():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "EUR", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    assert (len(user_transaction.objects) == 1)
    ut = repo.delete_transaction(ut.id)
    assert (len(user_transaction.objects) == 0)


def test_delete_transaction_when_exists_by_source_id():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "EUR", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    assert (len(user_transaction.objects) == 1)
    ut = repo.do_delete_transaction_by_source_id(source_id=ut.source_id)
    assert (len(user_transaction.objects) == 0)


def test_fetch_transaction():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    user_transaction.objects.all().delete()
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    ut = repo.fetch_transaction(ut.id)
    assert (ut.user_id == 1)
    assert (ut.symbol == "OXT")
    assert (ut.currency == "USD")
    assert (ut.operation == 'Added')
    assert (ut.source_id == ObjectId('666f6f2d6261722d71757578'))
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=None, operation='Added')
    ut = repo.fetch_transaction(ut.id)
    assert (ut.source_id is None)


def test_fetch_distinct_user_ids():
    config = configure_app()
    repo = TransactionRepository(config, mock_log)
    do_connect(config)
    user_transaction.objects.all().delete()
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    items =  ut = repo.do_fetch_distinct_user_ids()
    assert (len(items)==1)
    user_transaction.objects.all().delete()
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    ut = repo.insert_transaction(12, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    items =  ut = repo.do_fetch_distinct_user_ids()
    assert (len(items)==2)
    user_transaction.objects.all().delete()
    ut = repo.insert_transaction(1, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    ut = repo.insert_transaction(12, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    ut = repo.insert_transaction(3, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    ut = repo.insert_transaction(122, 1, 'OXT', 1, 1, "USD", "2020-01-01", "kraken",
                                 source_id=ObjectId('666f6f2d6261722d71757578'), operation='Added')
    items =  ut = repo.do_fetch_distinct_user_ids()
    assert (len(items)==4)